#include "../Bass.h"

#include <cstring>

#include <FileUtils.h>
#include <RandomUtils.h>
#include <logger/Log.h>

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include "../isc/ItemSetCollection.h"
#include "../itemstructs/Uint16ItemSet.h"
#include "../isc/ItemTranslator.h"

#include "DoubleDatabase.h"

DoubleDatabase::DoubleDatabase() : Database() {
	mDbLeft = mDbRight = NULL;
	mRowsLeft = mRowsRight = NULL;
}
DoubleDatabase::DoubleDatabase(Database *dbL, Database *dbR) {
	if(dbL->GetNumRows() != dbR->GetNumRows())
		THROW("Left and right databases must have the same number of rows.");
	if(dbL->GetDataType() != dbR->GetDataType())
		THROW("Left and right databases must have the same data type.");

	mDbLeft = dbL;
	mDbRight = dbR;

	// Copy things that can be copied
	mDbName = dbL->GetName() + "&" + dbR->GetName();
	mNumRows = dbL->GetNumRows();
	mNumTransactions = dbL->GetNumTransactions();
	mRowsLeft = dbL->GetRows();
	mRowsRight = dbR->GetRows();
	mNumItems = dbL->GetNumItems() + dbR->GetNumItems();
	mDataType = dbL->GetDataType();

	mHasBinSizes = dbL->HasBinSizes();
	mHasTransactionIds = dbL->HasTransactionIds();

	// Resize alphabets, shift right rows, and construct joint rows
	mAlphabetSize = dbL->GetAlphabetSize() + dbR->GetAlphabetSize();
	mAlphSplit = dbL->GetAlphabetSize();
	mRows = new ItemSet *[mNumRows];
	uint16 *tmpValues = new uint16[dbR->GetMaxSetLength()];
	uint32 tmpLen;
	for(uint32 r=0; r<mNumRows; r++) {
		// Resize left
		ItemSet *tmp = ItemSet::CreateEmpty(mDataType, mAlphabetSize, mRowsLeft[r]->GetUsageCount());
		tmp->Unite(mRowsLeft[r]);
		delete mRowsLeft[r];
		mRowsLeft[r] = tmp;

		// Resize and shift right
		tmpLen = mRowsRight[r]->GetLength();
		mRowsRight[r]->GetValuesIn(tmpValues);
		for(uint32 i=0; i<tmpLen; i++)
			tmpValues[i] += mAlphSplit;
		delete mRowsRight[r];
		mRowsRight[r] = ItemSet::Create(mDataType, tmpValues, tmpLen, mAlphabetSize, mRowsLeft[r]->GetUsageCount());

		// Construct joint rows
		mRows[r] = ItemSet::CreateEmpty(mDataType, mAlphabetSize, mRowsLeft[r]->GetUsageCount());
		mRows[r]->Unite(mRowsLeft[r]);
		mRows[r]->Unite(mRowsRight[r]);
	}
	delete[] tmpValues;

	// Essentials(tm) need to be re-worked
	mStdLaplace = false;
	mMaxSetLength = dbL->GetMaxSetLength() + dbR->GetMaxSetLength();
	SetAlphabetSize(mAlphabetSize);

	// Force alphabet upon its childs
	dbL->UseAlphabetFrom(this);
	dbR->UseAlphabetFrom(this);

	// Not per-se Essential (tm)
	mAlphabetDbOccs = NULL;
	mFastGenerateDBOccs = dbL->GetFastGenerateDBOccs();
	mItemTranslator = NULL;

	// Column Definition (c)
	mHasColumnDefinition = false;
	mNumColumns = 0;
	mColumnDefinition = new ItemSet*[mNumColumns];

	// Class Labels (.) -- take from left, as is default for most
	mNumClasses = dbL->GetNumClasses();
	mClassDefinition = new uint16[mNumClasses];
	memcpy(mClassDefinition, dbL->GetClassDefinition(), sizeof(uint16) * mNumClasses);
}
DoubleDatabase::DoubleDatabase(DoubleDatabase *ddb) : Database(ddb) {
	mDbLeft = ddb->mDbLeft->Clone();
	mDbRight = ddb->mDbRight->Clone();
	mRowsLeft = mDbLeft->GetRows();
	mRowsRight = mDbRight->GetRows();
}
DoubleDatabase::~DoubleDatabase() {
	delete mDbLeft;
	delete mDbRight;
}
DoubleDatabase* DoubleDatabase::Clone() {
	return new DoubleDatabase(this);
}
