#include "../global.h"

#include <time.h>
#include <Config.h>
#include <logger/Log.h>

#include <db/DoubleDatabase.h>
#include <isc/ItemSetCollection.h>

#include "../FicMain.h"
#include "../translate/TranslationRule.h"
#include "../translate/TranslationTable.h"

#include "GreedySearch.h"

GreedySearch::GreedySearch(Config *config) {
	mIscName = config->Read<string>("iscName");

	mFullLeftSet = NULL;
	mFullRightSet = NULL;
}
GreedySearch::~GreedySearch() {
	delete mFullLeftSet;
	delete mFullRightSet;
}
void GreedySearch::Init(DoubleDatabase *ddb) {
	RemodelSearch::Init(ddb);

	uint16 alphSplit = ddb->GetAlphabetSplit();
	uint16 alphSize = ddb->GetAlphabetSize();
	mFullLeftSet = ItemSet::CreateEmpty(ddb);
	mFullRightSet = ItemSet::CreateEmpty(ddb);
	for(uint16 i=0; i<alphSplit; i++)
		mFullLeftSet->AddItemToSet(i);
	for(uint16 i=alphSplit; i<alphSize; i++)
		mFullRightSet->AddItemToSet(i);
}

TranslationTable* GreedySearch::MineThatTable() {
	char temp[500]; string s;
	// Init file log
	OpenProgressFile();

	// Init TT 
	TranslationTable *tt = new TranslationTable(mDDB);
	uint32 numRows = mDDB->GetLeftDatabase()->GetNumRows();
	uint32 numIL = mDDB->GetAlphabetSplit();	
	uint32 numIR = mDDB->GetRightDatabase()->GetAlphabetSize();
	uint32 minSuppHalf = numRows / 2;
	tt->InitTranslation();
	
	// Init candidates
	bool freshlyMined;
	ItemSetCollection *isc = FicMain::ProvideItemSetCollection(mIscName, mDDB, freshlyMined, 1);
	uint64 numCands = isc->GetNumItemSets();
#if defined (_MSC_VER) && defined (_WINDOWS)
	sprintf_s(temp, 500, " # candidates:\t\t%I64d", numCands); s = temp; LOG_MSG(s);
#elif defined (__GNUC__) && defined (__unix__)
	sprintf_s(temp, 500, " # candidates:\t\t%lu", numCands); s = temp; LOG_MSG(s);
#endif
	
	time_t mSearchStart = time(NULL);

	// The Loop
	ItemSet *cand;
	uint32 numDiscarded = 0;
	double gain, gainBoth=0, gainToLeft=0, gainToRight=0;
	// double gainBothMS=0, gainToLeftMS=0, gainToRightMS=0;
	int startMs = getMilliCount();
	for(uint64 curCand=0; curCand<numCands; curCand++) {
		ProgressToScreen(curCand, numCands);
		cand = isc->GetNextItemSet();
		// Build the rule
		ItemSet *leftSet = cand->Remainder(mFullRightSet); // hypothesis: this is faster than intersection with opposite set
		ItemSet *rightSet = cand->Remainder(mFullLeftSet);
		delete cand;


		//printf("cand #%u\tsupp:%u, LSet: %u, RSet: %u\n", curCand, cand->GetSupport(), leftSet->GetLength(), rightSet->GetLength()); 
		if ( ( leftSet->IsEmpty() ||  rightSet->IsEmpty() ) && cand->GetSupport() < minSuppHalf) {
		  delete leftSet;
		  delete rightSet;
		  numDiscarded++;
		  continue;
		}
	      
		gainBoth = tt->ComputeGainForAddition(leftSet, rightSet, gainToRight, gainToLeft);
		// gainBothMS = tt->ComputeGainForAdditionMS(leftSet, rightSet, gainToRightMS, gainToLeftMS);
		// if (round(gainToLeft) != round(gainToLeftMS) ) {
		//    printf("LEFT %f vs. %f\n", gainToLeft, gainToLeftMS);
		// }
		// if ( round(gainToRight) != round(gainToRightMS) ) {
		//    printf("RIGHT %f vs. %f\n", gainToRight, gainToRightMS);
		// }
		if(gainBoth > 0.0 || gainToLeft > 0.0 || gainToRight > 0.0) {
		  TranslationRule *rule; 
		  if(gainToRight > gainBoth && gainToRight > gainToLeft) {
		    rule = new TranslationRule(leftSet, RuleToRight, rightSet);
		    gain = gainToRight;
		  } else if(gainToLeft > gainBoth) {
		    rule = new TranslationRule(leftSet, RuleToLeft, rightSet);
		    gain = gainToLeft;
		  } else { // gainToBoth is max
		    rule = new TranslationRule(leftSet, RuleBoth, rightSet);
		    gain = gainBoth;
		  }
		  tt->Add(rule); // single-pass; add immediately

		  ProgressToFileAddRule(rule, curCand, tt->GetCurStats(), getMilliSpan(startMs));
		  // -- Add pruning phase here? --
		  // Prune();

		} else {
		  delete leftSet;
		  delete rightSet;
		}
	}
	// Cleanup and roundup
	CloseProgressFile();
	delete isc;

	time_t seconds = time(NULL) - mSearchStart;
	sprintf_s(temp, 500, "** Searching %s with Greedy dataSplit=%u, dataSize=%ux%u, numCands=%u(-%u), searchTime=%d s", mIscName.c_str(),  numIL, numIR, numRows, numCands, numDiscarded, seconds); s = temp; LOG_MSG(s);

	return tt;
}
