#include "../global.h"

#include <time.h>

#include <Config.h>
#include <logger/Log.h>

#include <db/DoubleDatabase.h>
#include <isc/ItemSetCollection.h>

#include "../FicMain.h"
#include "../translate/CandidateRule.h"
#include "../translate/TranslationRule.h"
#include "../translate/TranslationTable.h"

#include "SelectSearch.h"

SelectSearch::SelectSearch(Config *config) {
	mIscName = config->Read<string>("iscName");
	mMaxPasses = config->Read<uint64>("maxPasses", 0);
	if (mMaxPasses == 0) mMaxPasses = UINT64_MAX_VALUE;
	mTopCands = config->Read<uint64>("topCands", 1);

	mFullLeftSet = NULL;
	mFullRightSet = NULL;
}
SelectSearch::~SelectSearch() {
	delete mFullLeftSet;
	delete mFullRightSet;
}
void SelectSearch::Init(DoubleDatabase *ddb) {
	RemodelSearch::Init(ddb);

	uint16 alphSplit = ddb->GetAlphabetSplit();
	uint16 alphSize = ddb->GetAlphabetSize();
	mFullLeftSet = ItemSet::CreateEmpty(ddb);
	mFullRightSet = ItemSet::CreateEmpty(ddb);
	for (uint16 i = 0; i<alphSplit; i++)
		mFullLeftSet->AddItemToSet(i);
	for (uint16 i = alphSplit; i<alphSize; i++)
		mFullRightSet->AddItemToSet(i);
}

TranslationTable* SelectSearch::MineThatTable() {
	// Init file log
	OpenProgressFile();

	// Init TT
	TranslationTable *tt = new TranslationTable(mDDB);
	uint32 numRows = mDDB->GetLeftDatabase()->GetNumRows();
	uint32 numIL = mDDB->GetAlphabetSplit();
	uint32 numIR = mDDB->GetRightDatabase()->GetAlphabetSize();
	uint32 minSuppHalf = numRows / 2;
	tt->InitTranslation();

	// Init candidates and cache them
	bool freshlyMined;
	ItemSetCollection *isc = FicMain::ProvideItemSetCollection(mIscName, mDDB, freshlyMined, 1);
	uint64 numCands = isc->GetNumItemSets();
	ItemSet **candCache = new ItemSet *[(uint32)numCands];
	uint32 numDiscarded = 0;
	time_t mSearchStart = time(NULL);
	for (uint32 c = 0; c<numCands; c++){ // attempt caching of all itemsets -- this may very well be impossible..
		// Build the rule
		candCache[c] = isc->GetNextItemSet();
		ItemSet *leftSet = candCache[c]->Remainder(mFullRightSet); // hypothesis: this is faster than intersection with opposite set
		ItemSet *rightSet = candCache[c]->Remainder(mFullLeftSet);
		//printf("cand #%u\tsupp:%u, LSet: %u, RSet: %u\n", curCand, cand->GetSupport(), leftSet->GetLength(), rightSet->GetLength()); 
		if ((leftSet->IsEmpty() || rightSet->IsEmpty()) && candCache[c]->GetSupport() < minSuppHalf) {
			delete candCache[c];
			candCache[c] = NULL;
			numDiscarded++;
		}
		delete leftSet;
		delete rightSet;
	}
	printf("---Discarded %u candidates\n", numDiscarded);

	// The Loop
	/// TODO: GROS TRAVAUX
	ItemSet *cand;
	double cMaxGain, gainBoth, gainToLeft, gainToRight;
	//double gainBothMS, gainToLeftMS, gainToRightMS;
	RuleDirection cRuleDirect;
	TranslationRule *maxRule;
	int numAdded = 0;
	list<CandidateRule*> bestRules;
	int numPasses = 0;
	CandidateRule **covers = new CandidateRule *[(uint32)mDDB->GetAlphabetSize()];
	do {
		int startMs = getMilliCount();
		for (uint64 ci = 0; ci<mDDB->GetAlphabetSize(); ci++) {
			covers[ci] = NULL;
		}

		for (uint64 curCand = 0; curCand<numCands; curCand++) {
			cand = candCache[curCand];
			if (cand == NULL)
				continue; // already added

			// Build the rule
			ItemSet *leftSet = cand->Remainder(mFullRightSet); // hypothesis: this is faster than intersection with opposite set
			ItemSet *rightSet = cand->Remainder(mFullLeftSet);

			// Compute gain and decide
			gainBoth = tt->ComputeGainForAddition(leftSet, rightSet, gainToRight, gainToLeft);
			if (gainToRight > gainBoth && gainToRight > gainToLeft) {
				cMaxGain = gainToRight;
				cRuleDirect = RuleToRight;
			}
			else if (gainToLeft > gainBoth) {
				cMaxGain = gainToLeft;
				cRuleDirect = RuleToLeft;
			}
			else {
				cMaxGain = gainBoth;
				cRuleDirect = RuleBoth;
			}

			if (cMaxGain <= 0.0) {
				delete leftSet;
				delete rightSet;
				continue;
			}

			int all_imp = 1;
			uint32 absId = mDDB->GetAlphabetSize() + 1;
			uint16 *valuesC = new uint16[mDDB->GetMaxSetLength()];
			cand->GetValuesIn(valuesC);
			for (uint32 i = 0; i < cand->GetLength(); i++){
				if (covers[valuesC[i]] != NULL && cMaxGain < covers[valuesC[i]]->GetGain()){
					valuesC[i] = absId;
					all_imp = 0;
				}
			}

			if (all_imp) {
				CandidateRule *maxCand = new CandidateRule(leftSet, cRuleDirect, rightSet, cMaxGain, curCand);
				for (uint32 i = 0; i<cand->GetLength(); i++){
					if (valuesC[i] != absId){
						covers[valuesC[i]] = maxCand;
					}
				}
			}
			delete[] valuesC;
		}

		numAdded = 0;
		// for (uint32 ci = 0 ; ci < mDDB->GetAlphabetSize(); ++ci ){
		//   if (covers[ci] != NULL)
		//     ProgressToFileCandidate(covers[ci], 0);
		// }
		for (uint32 ci = 0; ci < mDDB->GetAlphabetSize(); ++ci){
			if ((covers[ci] == NULL) || (candCache[covers[ci]->GetPos()] == NULL))
				continue;

			// Find position in ordered best rules
			candlist::iterator it = bestRules.begin();
			while (it != bestRules.end() && covers[ci]->GetGain() >(*it)->GetGain()){
				++it;
			}
			if (it != bestRules.begin()){
				if (it == bestRules.end()){
					bestRules.push_back(covers[ci]);
				}
				else {
					bestRules.insert(it, covers[ci]);
				}
				if (bestRules.size() > mTopCands) {
					bestRules.pop_front();
				}
			}
			else if (bestRules.size() < mTopCands) {
				bestRules.push_front(covers[ci]);
			}
			delete candCache[covers[ci]->GetPos()]; // remove candidate
			candCache[covers[ci]->GetPos()] = NULL;

		}
		numAdded = 0;
		list<CandidateRule*>::reverse_iterator rit;
		int spanMs = getMilliSpan(startMs);
		for (rit = bestRules.rbegin(); rit != bestRules.rend(); ++rit) {

			maxRule = new TranslationRule((*rit)->GetLeft()->Clone(), (*rit)->GetDirection(), (*rit)->GetRight()->Clone());

			tt->Add(maxRule);
			ProgressToFileAddRule(maxRule, (*rit)->GetPos(), tt->GetCurStats(), spanMs);

			// delete maxRule;
			numAdded++;
		}
		bestRules.clear();
		printf("PASS #%d\t# rules : %d/%u\n", numPasses, numAdded, tt->GetNumRules());
	} while (numAdded > 0.0 && ++numPasses < mMaxPasses);

	printf(" Stopped. # rules selected: %u\n", tt->GetNumRules());

	// Cleanup and roundup
	CloseProgressFile();
	/* for (uint32 ci = 0 ; ci < mDDB->GetAlphabetSize(); ++ci ){
	covers[ci] = NULL;
	}
	for(uint32 c=0; c<numCands; c++)
	delete candCache[c];
	delete[] candCache; */
	/*	for(uint64 ci=0; ci<mDDB->GetAlphabetSize(); ci++)
	delete covers[ci];
	delete[] covers;
	*/
	delete isc;

	char temp[500]; string s;
	time_t seconds = time(NULL) - mSearchStart;
	sprintf_s(temp, 500, "** Searching %s with Select dataSplit=%u, dataSize=%ux%u, numCands=%u(-%u), searchTime=%d s, maxPasses=%u, topCands=%u", mIscName.c_str(), numIL, numIR, numRows, numCands, numDiscarded, seconds, mMaxPasses, mTopCands); s = temp; LOG_MSG(s);

	return tt;
}
