#include "../global.h"

#include <logger/Log.h>
#include <db/DoubleDatabase.h>
#include <itemstructs/Uint16ItemSet.h>
#include "TranslationRule.h"

#include "TranslationTable.h"

TranslationTable::TranslationTable(DoubleDatabase *ddb) {
	mDDB = ddb;
	mRules = new rulelist(); // start with empty rulelist
	mItems = NULL;	
	mVerticalDb = NULL;
	mLeftItems = NULL;
	mRightItems = NULL;

	mLeftItemsLast = NULL;
	mRightItemsLast = NULL;
	
	mNumRows= 0;
	mUncovered = NULL;
	mErrors = NULL;
	mColUncoveredCounts = NULL;
	mColErrorCounts = NULL;
	mGainCounts = NULL;
	mZeroCounts = NULL;
	mAlphSize = mAlphSplit = 0;

	mTmpValues = new uint16[mDDB->GetMaxSetLength()];
	mLeftStdLens = mDDB->GetLeftDatabase()->GetStdLengths();
	mRightStdLens = mDDB->GetRightDatabase()->GetStdLengths();
}
TranslationTable::TranslationTable(const TranslationTable &tt) {
        mItems = NULL;	
	mVerticalDb = NULL;
		
        mDDB = tt.mDDB;
	mRules = new rulelist();

	mLeftItems = tt.mLeftItems->Clone();
	mRightItems = tt.mRightItems->Clone();

	mLeftItemsLast = tt.mLeftItemsLast->Clone();
	mRightItemsLast = tt.mRightItemsLast->Clone();

	// Duplicate rules
	rulelist::iterator rit=tt.mRules->begin(), rend=tt.mRules->end();
	for( ; rit != rend; ++rit)
		mRules->push_back(((TranslationRule*)(*rit))->Clone());

	mNumRows = 0;
	mErrors = NULL;
	mColUncoveredCounts = NULL;
	mColErrorCounts = NULL;
	mGainCounts = NULL;
	mZeroCounts = NULL;
	mAlphSize = mAlphSplit = 0;

	mLeftItems = NULL;
	mRightItems = NULL;

	mTmpValues = new uint16[mDDB->GetMaxSetLength()];
	mLeftStdLens = mDDB->GetLeftDatabase()->GetStdLengths();
	mRightStdLens = mDDB->GetRightDatabase()->GetStdLengths();
}

TranslationTable::~TranslationTable() {
	rulelist::iterator rit=mRules->begin(), rend=mRules->end();
	for( ; rit != rend; ++rit)
		delete (*rit);
	delete mRules;

	for(uint32 r=0; r<mNumRows; r++) {
		delete mUncovered[r];
		delete mErrors[r];
	}
	delete[] mUncovered;
	delete[] mErrors;
	delete[] mColUncoveredCounts;
	delete[] mColErrorCounts;
	delete[] mGainCounts;
	delete[] mZeroCounts;
	delete mLeftItems;
	delete mRightItems;
	delete mLeftItemsLast;
	delete mRightItemsLast;

	delete[] mTmpValues;
	
	delete mVerticalDb;	

}
TranslationTable* TranslationTable::Clone() {
	return new TranslationTable(*this);
}
void TranslationTable::InitTranslation() {

	mDDB->ComputeEssentials();
	
	mLeftItems = ItemSet::CreateEmpty(mDDB);
	mRightItems = ItemSet::CreateEmpty(mDDB);

	mLeftItemsLast = ItemSet::CreateEmpty(mDDB);
	mRightItemsLast = ItemSet::CreateEmpty(mDDB);

	mNumRows = mDDB->GetNumRows();
	mAlphSize = mDDB->GetAlphabetSize();
	mAlphSplit = mDDB->GetAlphabetSplit();

	// CorrectionTable = Uncovered \cup Errors
	mUncovered = new ItemSet *[mNumRows];
	mErrors = new ItemSet *[mNumRows];
	for(uint32 r=0; r<mNumRows; r++) {
	  mUncovered[r] = mDDB->GetRow(r)->Clone(); // start with complete row uncovered
	  mErrors[r] = ItemSet::CreateEmpty(mDDB);	// start without errors
	}

	// Init counts for easy encoded size computation
	mColUncoveredCounts = new uint32[mAlphSize];
	mColErrorCounts = new uint32[mAlphSize];
	mGainCounts = new uint32[mAlphSize];
	mZeroCounts = new uint32[mAlphSize];
	for(uint32 i=0; i<mAlphSize; i++)
		mZeroCounts[i] = 0;

	// Compute initial sizes
	ComputeEncodedSize();
}

void TranslationTable::ClearLast() {
  delete mLeftItemsLast;
  delete mRightItemsLast;
  mLeftItemsLast = ItemSet::CreateEmpty(mDDB);
  mRightItemsLast = ItemSet::CreateEmpty(mDDB);
}
void TranslationTable::Add(TranslationRule *rule) {
	rule->GetLeft()->ComputeStandardLength(mDDB->GetLeftDatabase()->GetStdLengths());
	rule->GetRight()->ComputeStandardLength(mDDB->GetRightDatabase()->GetStdLengths());
	mRules->push_back(rule);
	ApplyRule(rule);
	ComputeEncodedSize();

	mLeftItems->Unite(rule->GetLeft());
	mRightItems->Unite(rule->GetRight());

	mLeftItemsLast->Unite(rule->GetLeft());
	mRightItemsLast->Unite(rule->GetRight());

	/// NO NEED TO EMPTY SUPPORTS STORES
	// mSetsLMap.clear();
	// mSetsRMap.clear();

}
double TranslationTable::ComputeGainForAddition(TranslationRule *rule) {
	rule->GetLeft()->ComputeStandardLength(mDDB->GetLeftDatabase()->GetStdLengths());
	rule->GetRight()->ComputeStandardLength(mDDB->GetRightDatabase()->GetStdLengths());
	double gain = 0.0;
	THROW_NOT_IMPLEMENTED(); // TODO
	return gain;
}
double TranslationTable::ComputeGainForAdditionMS(ItemSet *leftSet, ItemSet *rightSet, double &toRight, double &toLeft) {
	// Start with zeroes and obtain access to some required structures
	toRight = toLeft = 0.0;
	ItemSet **leftRows = mDDB->GetLeftRows();
	ItemSet **rightRows = mDDB->GetRightRows();
	double *leftStdLens = mDDB->GetLeftDatabase()->GetStdLengths();
	double *rightStdLens = mDDB->GetRightDatabase()->GetStdLengths();
	uint16 *values = new uint16[mDDB->GetMaxSetLength()];
	uint32 len;

	// Compute would-be effect given current correction table
	for(uint32 r=0; r<mNumRows; r++) {
		if(leftSet->IsEmpty() || leftRows[r]->IsSubset(leftSet)) { // l->r matches
			if(!rightSet->IsEmpty()) { // empty? no influence
				// The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
				ItemSet *cover = mUncovered[r]->Intersection(rightSet);
				cover->GetValuesIn(values); len = cover->GetLength();
				for(uint32 i=0; i<len; i++)
					toRight += rightStdLens[values[i]];
				delete cover;

				// Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
				ItemSet *errors = rightSet->Remainder(rightRows[r]);
				errors->Remove(mErrors[r]);
				errors->GetValuesIn(values); len = errors->GetLength();
				for(uint32 i=0; i<len; i++)
					toRight -= rightStdLens[values[i]];
				// printf("A2+\n");
				// printf("%d: %s (%s)\n", r, rightRows[r]->ToString().c_str(), rightSet->ToString().c_str());
				delete errors;
				// printf("A2-\n");
				
			}
		}
		if(rightSet->IsEmpty() || rightRows[r]->IsSubset(rightSet)) { // l<-r matches
			if(!leftSet->IsEmpty()) { // empty? no influence
				// The intersection between uncovered and leftset results in a smaller correction table and thus positive gain
				ItemSet *cover = mUncovered[r]->Intersection(leftSet);
				cover->GetValuesIn(values); len = cover->GetLength();
				for(uint32 i=0; i<len; i++)
					toLeft += leftStdLens[values[i]];
				delete cover;

				// Everything in leftset but not in the original row or current errors results in a larger correction table and thus negative gain
				ItemSet *errors = leftSet->Remainder(leftRows[r]);
				errors->Remove(mErrors[r]);
				errors->GetValuesIn(values); len = errors->GetLength();
				for(uint32 i=0; i<len; i++)
					toLeft -= leftStdLens[values[i]];
				delete errors;
			}
		}
	}

	// Cleanup
	delete[] values;

	// End.
	double ruleStdSize = leftSet->ComputeStandardLength(leftStdLens) + rightSet->ComputeStandardLength(rightStdLens);
	double gain = toLeft + toRight - (ruleStdSize + 1.0);
	toLeft -= (ruleStdSize + 2.0); // extra bit penalty for direction
	toRight -= (ruleStdSize + 2.0); // extra bit penalty for direction
	return gain;
}

// ########################################
double TranslationTable::ComputeGainForAddition(ItemSet *leftSet, ItemSet *rightSet, double &toRight, double &toLeft) {
  
         uint16 alphSplit = mDDB->GetAlphabetSplit();
	 uint16 alphSize = mDDB->GetAlphabetSize();

	 if( mItems == NULL){
	   // Construct vertical db, ie a tidset per singleton
	   mVerticalDb = mDDB->Transpose();
	   mItems = (Uint16ItemSet**)(mVerticalDb->GetRows());	// each entry corresponds to an uint16 itemset -> the tidset
	   alphabet *alph = mDDB->GetAlphabet();
	   alphabet::iterator it=alph->begin();
	   double *stdLensL = mDDB->GetLeftDatabase()->GetStdLengths();
	   double *stdLensR = mDDB->GetRightDatabase()->GetStdLengths();
	   for(uint32 i=0; i<alphSize; i++) {  // foreach item
	     mItems[i]->SetID(i);
	     // set its support (note: this is actually equivalent to the length of the tidset)
	     mItems[i]->SetSupport(it->second);
	     // set its singleton standard code length
	     mItems[i]->SetStandardLength(i<alphSplit ? stdLensL[i] : stdLensR[i]);
	     it++;
	   }
	  }

	  Uint16ItemSet *leftTidSet, *rightTidSet;
	  toRight = toLeft = 0.0;
	  std::map<string, Uint16ItemSet*>::iterator it;

	  // ### RETRIEVING LHS
	  uint16 *valitems = new uint16[alphSize];
	  it = mSetsLMap.find(leftSet->ToString(0,0).c_str());
	  if ( it == mSetsLMap.end() ) {
	    if ( ! leftSet->IsEmpty() ){
	      leftSet->GetValuesIn(valitems);
	      leftTidSet = mItems[valitems[0]]->Clone();
	      for(uint16 ii=1; ii<leftSet->GetLength(); ii++) {
		Uint16ItemSet *tidSet = (Uint16ItemSet*)(leftTidSet->Intersection(mItems[valitems[ii]]));
		delete leftTidSet;
		leftTidSet = tidSet;							
	      }
	      mSetsLMap[leftSet->ToString(0,0).c_str()] = leftTidSet;
	    }
	  }
	  else {
	    leftTidSet = it->second;
	  }
	  
	  // ### RETRIEVING RHS
	  it = mSetsRMap.find(rightSet->ToString(0,0).c_str());
	  if ( it == mSetsRMap.end() ) {
	    if ( ! rightSet->IsEmpty() ){
	      rightSet->GetValuesIn(valitems);
	      rightTidSet = mItems[valitems[0]]->Clone();
	      for(uint16 ii=1; ii<rightSet->GetLength(); ii++) {
		Uint16ItemSet *tidSet = (Uint16ItemSet*)(rightTidSet->Intersection(mItems[valitems[ii]]));
		delete rightTidSet;
		rightTidSet = tidSet;
	      }
	      mSetsRMap[rightSet->ToString(0,0).c_str()] = rightTidSet;
	    }
	  }
	  else {
	    rightTidSet = it->second;
	  }

	  //////////  COMPUTE INTERSECTION
	  Uint16ItemSet *alpha, *beta;
	  uint32 alphaL=0, betaL=0, gammaL=0;
	  double *leftStdLens = mDDB->GetLeftDatabase()->GetStdLengths();
	  double *rightStdLens = mDDB->GetRightDatabase()->GetStdLengths();
	  double stdLLeft = leftSet->ComputeStandardLength(leftStdLens);
	  double stdLRight = rightSet->ComputeStandardLength(rightStdLens); 

	  if ( rightSet->IsEmpty() ){
	    beta = new Uint16ItemSet(mNumRows);
	    beta->FillHerUp(mNumRows);
	    betaL = mNumRows;
	    if ( ! leftSet->Intersects(mLeftItems) ){
	      toLeft += stdLLeft*leftTidSet->GetLength();
	      beta->Remove(leftTidSet);
	      betaL = beta->GetLength();
	    }
	  }
	  else if ( leftSet->IsEmpty() ){
	    alpha = new Uint16ItemSet(mNumRows);
	    alpha->FillHerUp(mNumRows);
	    alphaL = mNumRows;
	    if ( ! rightSet->Intersects(mRightItems) ){
	      toRight += stdLRight*rightTidSet->GetLength();
	      alpha->Remove(rightTidSet);
	      alphaL = alpha->GetLength();
	    }
	  }
	  else {
	    if ( rightSet->Intersects(mRightItems) ){
	      alpha = leftTidSet->Clone();
	      alphaL = alpha->GetLength();
	    }
	    else {
	      alpha = leftTidSet->Clone();
	      alpha->Remove(rightTidSet);
	      alphaL = alpha->GetLength();
	      toRight += stdLRight*(leftTidSet->GetLength() - alphaL);
	    } 
	    if ( leftSet->Intersects(mLeftItems) ){
	      beta = rightTidSet->Clone();	    
	      betaL = beta->GetLength();
	    }
	    else {
	      beta = rightTidSet->Clone();
	      beta->Remove(leftTidSet);
	      betaL = beta->GetLength();
	      toLeft += stdLLeft*(rightTidSet->GetLength() - betaL);
	    }
	  }

	// Start with zeroes and obtain access to some required structures
	ItemSet **leftRows = mDDB->GetLeftRows();
	ItemSet **rightRows = mDDB->GetRightRows();
	uint16 *values = new uint16[mDDB->GetMaxSetLength()];
	uint16 *rowsIds = new uint16[mNumRows];
	uint32 len;
	
	/// ALPHA
	if( alphaL > 0 ) { // empty? no influence
	  alpha->GetValuesIn(rowsIds);
	  for(uint32 ri=0; ri<alphaL; ri++) {
	    // The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
	    ItemSet *cover = mUncovered[rowsIds[ri]]->Intersection(rightSet);
	    cover->GetValuesIn(values); len = cover->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toRight += rightStdLens[values[i]];
	    //printf("A1+\n");
	    delete cover;
	    //printf("A1-\n");
	  
	    // Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
	    ItemSet *errors = rightSet->Remainder(rightRows[rowsIds[ri]]);
	    errors->Remove(mErrors[rowsIds[ri]]);
	    errors->GetValuesIn(values); len = errors->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toRight -= rightStdLens[values[i]];
	    //printf("A2+\n");
	    delete errors;
	    //printf("A2-\n");
	  }
	}

	/// BETA
	if( betaL > 0 ) { // empty? no influence
	  beta->GetValuesIn(rowsIds);
	  for(uint32 ri=0; ri<betaL; ri++) {
	    // The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
	    ItemSet *cover = mUncovered[rowsIds[ri]]->Intersection(leftSet);
	    cover->GetValuesIn(values); len = cover->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toLeft += leftStdLens[values[i]];
	    //printf("B1+\n");
	    delete cover;
	    //printf("B1-\n");
	  
	    // Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
	    ItemSet *errors = leftSet->Remainder(leftRows[rowsIds[ri]]);
	    errors->Remove(mErrors[rowsIds[ri]]);
	    errors->GetValuesIn(values); len = errors->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toLeft -= leftStdLens[values[i]];
	    //printf("B2+\n");
	    delete errors;
	    //printf("B2-\n");
	  }
	}

	// Cleanup
	delete[] values;
	delete[] rowsIds;
	if ( ! rightSet->IsEmpty() )
	  delete alpha;
	if ( ! leftSet->IsEmpty() )
	  delete beta;

	// End.
	double ruleStdSize = stdLRight + stdLLeft;
	double gain = toLeft + toRight - (ruleStdSize + 1.0);
	toLeft -= (ruleStdSize + 2.0); // extra bit penalty for direction
	toRight -= (ruleStdSize + 2.0); // extra bit penalty for direction
	return gain;


}
  
double TranslationTable::ComputeGainForAdditionMod(ItemSet *leftSet, ItemSet *rightSet, ItemSet *leftTidSet, ItemSet *rightTidSet, double &toRight, double &toLeft) {
	// Start with zeroes and obtain access to some required structures
	toRight = toLeft = 0.0;
	ItemSet **leftRows = mDDB->GetLeftRows();
	ItemSet **rightRows = mDDB->GetRightRows();
	double *leftStdLens = mDDB->GetLeftDatabase()->GetStdLengths();
	double *rightStdLens = mDDB->GetRightDatabase()->GetStdLengths();
	uint16 *values = new uint16[mDDB->GetMaxSetLength()];
	uint32 len;
	
	double stdLLeft = leftSet->ComputeStandardLength(leftStdLens);
	double stdLRight = rightSet->ComputeStandardLength(rightStdLens);

	ItemSet *alpha = leftTidSet->Remainder(rightTidSet);
	ItemSet *beta = rightTidSet->Remainder(leftTidSet);
	uint32 alphaL = alpha->GetLength();
	uint32 betaL = beta->GetLength();
	uint32 gammaL = leftTidSet->GetLength() - alphaL;

	/// ALPHA
	if(!rightSet->IsEmpty()) { // empty? no influence
	  uint16 *alphaV = new uint16[alphaL];
	  alpha->GetValuesIn(alphaV);

	  for(uint32 ri=0; ri<alphaL; ri++) {
	    uint32 r = alphaV[ri];
	    // The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
	    ItemSet *cover = mUncovered[r]->Intersection(rightSet);
	    cover->GetValuesIn(values); len = cover->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toRight += rightStdLens[values[i]];
	    delete cover;
	  
	    // Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
	    ItemSet *errors = rightSet->Remainder(rightRows[r]);
	    errors->Remove(mErrors[r]);
	    errors->GetValuesIn(values); len = errors->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toRight -= rightStdLens[values[i]];
	    delete errors;
	  }
	  delete[] alphaV; 
	}

	/// BETA
	if(!leftSet->IsEmpty()) { // empty? no influence
	  uint16 *betaV = new uint16[betaL];
	  beta->GetValuesIn(betaV);

	  for(uint32 ri=0; ri<betaL; ri++) {
	    uint32 r = betaV[ri];
	    // The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
	    ItemSet *cover = mUncovered[r]->Intersection(leftSet);
	    cover->GetValuesIn(values); len = cover->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toLeft += leftStdLens[values[i]];
	    delete cover;
	  
	    // Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
	    ItemSet *errors = leftSet->Remainder(leftRows[r]);
	    errors->Remove(mErrors[r]);
	    errors->GetValuesIn(values); len = errors->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toLeft -= leftStdLens[values[i]];
	    delete errors;
	  }
	  delete[] betaV; 
	}

	/// GAMMA
	toRight += stdLRight*gammaL;
	toLeft += stdLLeft*gammaL;

	// Cleanup
	delete[] values;

	// End.
	double ruleStdSize = stdLRight + stdLLeft;
	double gain = toLeft + toRight - (ruleStdSize + 1.0);
	toLeft -= (ruleStdSize + 2.0); // extra bit penalty for direction
	toRight -= (ruleStdSize + 2.0); // extra bit penalty for direction
	return gain;
}
// ##################################################

void TranslationTable::ComputeGainForAdditionFromTree(ItemSet *leftSet, ItemSet *rightSet, Uint16ItemSet *leftTidSet, Uint16ItemSet *rightTidSet, double &toRight, double &toLeft) {
	  toRight = toLeft = 0.0;
	  //////////  COMPUTE INTERSECTION
	  Uint16ItemSet *alpha, *beta;
	  uint32 alphaL=0, betaL=0, gammaL=0;
	  double *leftStdLens = mDDB->GetLeftDatabase()->GetStdLengths();
	  double *rightStdLens = mDDB->GetRightDatabase()->GetStdLengths();
	  double stdLLeft = leftSet->ComputeStandardLength(leftStdLens);
	  double stdLRight = rightSet->ComputeStandardLength(rightStdLens); 
	  
	  if ( rightSet->Intersects(mRightItems) ){
	    alpha = leftTidSet->Clone();
	    alphaL = alpha->GetLength();
	  }
	  else {
	    alpha = leftTidSet->Clone();
	    alpha->Remove(rightTidSet);
	    alphaL = alpha->GetLength();
	    toRight += stdLRight*(leftTidSet->GetLength() - alphaL);
	  } 

	// Start with zeroes and obtain access to some required structures
	ItemSet **leftRows = mDDB->GetLeftRows();
	ItemSet **rightRows = mDDB->GetRightRows();
	uint16 *values = new uint16[mDDB->GetMaxSetLength()];
	uint16 *rowsIds = new uint16[mNumRows];
	uint32 len;
	
	/// ALPHA
	if( alphaL > 0 ) { // empty? no influence
	  alpha->GetValuesIn(rowsIds);
	  for(uint32 ri=0; ri<alphaL; ri++) {
	    // The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
	    ItemSet *cover = mUncovered[rowsIds[ri]]->Intersection(rightSet);
	    cover->GetValuesIn(values); len = cover->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toRight += rightStdLens[values[i]];
	    delete cover;
	  
	    // Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
	    ItemSet *errors = rightSet->Remainder(rightRows[rowsIds[ri]]);
	    errors->Remove(mErrors[rowsIds[ri]]);
	    errors->GetValuesIn(values); len = errors->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toRight -= rightStdLens[values[i]];
	    delete errors;
	  }
	}

	  if ( leftSet->Intersects(mLeftItems) ){
	    beta = rightTidSet->Clone();	    
	    betaL = beta->GetLength();
	  }
	  else {
	    beta = rightTidSet->Clone();
	    beta->Remove(leftTidSet);
	    betaL = beta->GetLength();
	    toLeft += stdLLeft*(rightTidSet->GetLength() - betaL);
 	  }

	/// BETA
	if( betaL > 0 ) { // empty? no influence
	  beta->GetValuesIn(rowsIds);
	  for(uint32 ri=0; ri<betaL; ri++) {
	    // The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
	    ItemSet *cover = mUncovered[rowsIds[ri]]->Intersection(leftSet);
	    cover->GetValuesIn(values); len = cover->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toLeft += mLeftStdLens[values[i]];
	    delete cover;
	  
	    // Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
	    ItemSet *errors = leftSet->Remainder(leftRows[rowsIds[ri]]);
	    errors->Remove(mErrors[rowsIds[ri]]);
	    errors->GetValuesIn(values); len = errors->GetLength();
	    for(uint32 i=0; i<len; i++)
	      toLeft -= mLeftStdLens[values[i]];
	    delete errors;
	  }
	}

	// Cleanup
	delete[] values;
	delete[] rowsIds;
	delete alpha;
	delete beta;

	// End.
}


void TranslationTable::ComputeGainForAdditionFromTreeMS(ItemSet *leftSet, ItemSet *rightSet, Uint16ItemSet *leftTidSet, Uint16ItemSet *rightTidSet, double &toRight, double &toLeft) {
	// Start with zeroes and obtain access to some required structures
	toRight = toLeft = 0.0;
	ItemSet **leftRows = mDDB->GetLeftRows();
	ItemSet **rightRows = mDDB->GetRightRows();
	uint32 len;

	// -- Compute would-be effect given current correction table
	// Left to right
	uint16 *rows = leftTidSet->GetSet();
	uint32 numRows = leftTidSet->GetLength();
	for(uint32 r=0; r<numRows; r++) {
		// The intersection between uncovered and rightset results in a smaller correction table and thus positive gain
		ItemSet *cover = mUncovered[rows[r]]->Intersection(rightSet);
		cover->GetValuesIn(mTmpValues); len = cover->GetLength();
		for(uint32 i=0; i<len; i++)
			toRight += mRightStdLens[mTmpValues[i]];
		delete cover;

		// Everything in rightset but not in the original row or current errors results in a larger correction table and thus negative gain
		ItemSet *errors = rightSet->Remainder(rightRows[rows[r]]);
		errors->Remove(mErrors[rows[r]]);
		errors->GetValuesIn(mTmpValues); len = errors->GetLength();
		for(uint32 i=0; i<len; i++)
			toRight -= mRightStdLens[mTmpValues[i]];
		delete errors;
	}

	// Right to left
	rows = rightTidSet->GetSet();
	numRows = rightTidSet->GetLength();
	for(uint32 r=0; r<numRows; r++) {
		// The intersection between uncovered and leftset results in a smaller correction table and thus positive gain
		ItemSet *cover = mUncovered[rows[r]]->Intersection(leftSet);
		cover->GetValuesIn(mTmpValues); len = cover->GetLength();
		for(uint32 i=0; i<len; i++)
			toLeft += mLeftStdLens[mTmpValues[i]];
		delete cover;

		// Everything in leftset but not in the original row or current errors results in a larger correction table and thus negative gain
		ItemSet *errors = leftSet->Remainder(leftRows[rows[r]]);
		errors->Remove(mErrors[rows[r]]);
		errors->GetValuesIn(mTmpValues); len = errors->GetLength();
		for(uint32 i=0; i<len; i++)
			toLeft -= mLeftStdLens[mTmpValues[i]];
		delete errors;
	}
}
float** TranslationTable::ComputeRowGainPotentials() {
	double *stdLensL = mDDB->GetLeftDatabase()->GetStdLengths();
	double *stdLensR = mDDB->GetRightDatabase()->GetStdLengths();

	float **potentials = new float *[2];
	potentials[LEFT] = new float[mNumRows];
	potentials[RIGHT] = new float[mNumRows];

	for(uint32 r=0; r<mNumRows; r++) {
		potentials[LEFT][r] = potentials[RIGHT][r] = 0.0f;
		mUncovered[r]->GetValuesIn(mTmpValues);
		for(uint32 i=0; i<mUncovered[r]->GetLength(); i++) {
			if(mTmpValues[i] < mAlphSplit)
				potentials[LEFT][r] += (float)stdLensL[mTmpValues[i]];
			else
				potentials[RIGHT][r] += (float)stdLensR[mTmpValues[i]];
		}
	}

	return potentials;
}

void TranslationTable::ComputeEncodedSize() {
	// Re-start from zero (could/should be improved later--delta counting only)
	memcpy_s(mColUncoveredCounts, mAlphSize*sizeof(uint32), mZeroCounts, mAlphSize*sizeof(uint32));
	memcpy_s(mColErrorCounts, mAlphSize*sizeof(uint32), mZeroCounts, mAlphSize*sizeof(uint32));
	mCurStats.encLeftSize = mCurStats.encRightSize = mCurStats.encSize = mCurStats.encTableSize = 0.0;
	mCurStats.nbUncoveredR = mCurStats.nbErrorsR = mCurStats.nbUncoveredL = mCurStats.nbErrorsL = 0;

	mCurStats.numRules = mRules->size();

	// For each row
	uint16 *values = new uint16[mDDB->GetAlphabetSize()];
	for(uint32 r=0; r<mNumRows; r++) {
		mUncovered[r]->GetValuesIn(values);
		for(uint32 i=0; i<mUncovered[r]->GetLength(); i++)
			mColUncoveredCounts[values[i]]++;
		mErrors[r]->GetValuesIn(values);
		for(uint32 i=0; i<mErrors[r]->GetLength(); i++)
			mColErrorCounts[values[i]]++;
	}
	delete[] values;

	// Compute encoded data sizes
	double *stdLens = mDDB->GetLeftDatabase()->GetStdLengths();

	for(uint32 i=0; i<mAlphSplit; i++) {
	  mCurStats.encLeftSize += (mColUncoveredCounts[i]+mColErrorCounts[i]) * stdLens[i];
	  mCurStats.nbUncoveredL += mColUncoveredCounts[i];
	  mCurStats.nbErrorsL += mColErrorCounts[i];
	  }
	stdLens = mDDB->GetRightDatabase()->GetStdLengths();
	for(uint32 i=mAlphSplit; i<mAlphSize; i++){
	  mCurStats.encRightSize += (mColUncoveredCounts[i]+mColErrorCounts[i]) * stdLens[i];
	  mCurStats.nbUncoveredR += mColUncoveredCounts[i];
	  mCurStats.nbErrorsR += mColErrorCounts[i];
	  }

	// For each rule
	rulelist::iterator rit=mRules->begin(), rend=mRules->end();
	for( ; rit != rend; ++rit)
		mCurStats.encTableSize += (*rit)->GetEncodedSize();

	mCurStats.encSize = mCurStats.encLeftSize + mCurStats.encRightSize + mCurStats.encTableSize;
}
void TranslationTable::ResetTranslation() {
	// Reset the correction table
	for(uint32 r=0; r<mNumRows; r++) {
		mUncovered[r]->Unite(mDDB->GetRow(r));	// re-start with complete row uncovered
		mErrors[r]->CleanTheSlate();			// re-start without errors
	}
	// Correction counts are currently reset upon each ComputeEncodedSize() -- no need to do that here
}
void TranslationTable::PerformTranslation(const bool ltr, const bool rtl) {
	if(mNumRows==0)
		throw string("Perform InitTranslation() prior to translation.");

	// Reset all counts
	ResetTranslation();

	// Apply each rule (in rule list ordering)
	rulelist::iterator rit=mRules->begin(), rend=mRules->end();
	for( ; rit != rend; ++rit)
		ApplyRule(*rit, ltr, rtl);

	ComputeEncodedSize();
}
void TranslationTable::ApplyRule(TranslationRule *rule, const bool ltr /*= true*/, const bool rtl /*= true*/) {
	if(ltr && rule->GetDirection()!=RuleToLeft) {	 // Attempt to apply rule left to right on each row
		ItemSet *antecedent = rule->GetLeft();
		ItemSet *consequent = rule->GetRight();

		ApplyRule(antecedent, mDDB->GetLeftRows(), consequent, mDDB->GetRightRows());
	}
	if(rtl && rule->GetDirection()!=RuleToRight) {	 // Attempt to apply rule right to left on each row
		ItemSet *antecedent = rule->GetRight();
		ItemSet *consequent = rule->GetLeft();

		ApplyRule(antecedent, mDDB->GetRightRows(), consequent, mDDB->GetLeftRows());
	}
}
void TranslationTable::ApplyRule(ItemSet *antecedent, ItemSet **antRows, ItemSet *consequent, ItemSet **consRows) {
	for(uint32 r=0; r<mNumRows; r++) {
		if(antecedent->IsEmpty() || antRows[r]->IsSubset(antecedent)) { // Antecedent matches -> apply rule to this row (note: always applies rules with empty antecedent)
			if(!consequent->IsEmpty()) { // empty? no influence
				// the following part may have to be re-implemented later to speed things up
				mUncovered[r]->Remove(consequent); // cover everything in consequent
				ItemSet *errs = consequent->Remainder(consRows[r]); // everything in consequent but not in the full row is an error
				mErrors[r]->Unite(errs); // add errors to existing errors
				delete errs;
			}
		}
	}
}
void TranslationTable::WriteToDisk(const string &filename) {
	FILE *file;
	if(fopen_s(&file, filename.c_str(), "w") != 0) {
		printf("Cannot open TranslationTable file for writing!\nFilename: %s\n", filename.c_str());
		if(fopen_s(&file, filename.c_str(), "w") != 0) {
			printf("Second try failed as well..\n");
			return;
		}
	}
	rulelist::iterator rit,rend=mRules->end();

	// Determine max set lengths
	uint32 maxLenL = 0, maxLenR = 0;
	for(rit = mRules->begin(); rit != rend; ++rit) {
		if((*rit)->GetLeft()->GetLength() > maxLenL)
			maxLenL = (*rit)->GetLeft()->GetLength();
		if((*rit)->GetRight()->GetLength() > maxLenR)
			maxLenR = (*rit)->GetRight()->GetLength();
	}

	// Header
	fprintf(file, "fictt-1.0\n%u %u %u %u %u\n", mRules->size(), mAlphSplit, mAlphSize, maxLenL, maxLenR);

	// Rules
	for (rit = mRules->begin(); rit!=rend; ++rit)
		fprintf(file, "%s\n", (*rit)->ToString().c_str());

	// The end
	fclose(file);
}
void TranslationTable::ReadFromDisk(const string &filename)  {
	// Init
	char *buffer = new char[TTFILE_BUFFER_LENGTH];
	FILE *file = fopen(filename.c_str(), "r");
	if(file == NULL || ferror(file))
		THROW("Could not open TranslationTable for reading.");

	// Read header
	if(fgets(buffer, TTFILE_BUFFER_LENGTH, file) == NULL)
		THROW("Could not read header from file.");
	if(strcmp(buffer, "fictt-1.0\n") != 0)
		THROW("Unknown file format.");
	uint32 numRules, maxLenL, maxLenR;
	if(fscanf(file, "%u %u %u %u %u\n", &numRules, &mAlphSplit, &mAlphSize, &maxLenL, &maxLenR) < 5)
		THROW("Could not find appropiate fictt-1.0 header.");
	if(mAlphSize != mDDB->GetAlphabetSize() || mAlphSplit != mDDB->GetAlphabetSplit())
		THROW("Alphabet sizes do not match those of the database.");

	// Prepare rule reading
	uint16 *setL = new uint16[maxLenL];
	uint16 *setR = new uint16[maxLenR];
	uint32 setLenL, setLenR;
	ItemSet *isL, *isR;
	RuleDirection dir;

	// Read the given number of rules
	for(uint32 r=0; r<numRules; r++) {
		setLenL = setLenR = 0;
		dir = RuleBoth;

		// Read the sets and direction
		if(fgets(buffer, TTFILE_BUFFER_LENGTH, file) == NULL)
			THROW("No rule no table.");
		uint32 curLineLen = (uint32) strlen(buffer);
		if(curLineLen <= 6)
			THROW("Too short to be a rule.");

		// Count items
		char *bp = buffer;
		uint32 col=0;
		for(; col<curLineLen; col++) { // left
			if(*bp == '-')
				break;
			else if(*bp == ' ')
				setLenL++;
			bp++;
		}
		for(; col<curLineLen; col++) { // right (note: each item should be followed by a space, hence each rule/line should end with a space)
			if(*bp == ' ')
				setLenR++;
			bp++;
		}

		// Read left itemset
		bp = buffer;
		double stdLen = 0.0;
		double *stdLengths = mDDB->GetLeftDatabase()->GetStdLengths();
		for(uint32 i=0; i<setLenL; i++) {
			setL[i] = atoi(bp);
			stdLen += stdLengths[setL[i]];
			while(*bp != ' ')
				++bp;
			++bp; // skip space
		}
		isL = ItemSet::Create(mDDB->GetDataType(), setL, setLenL, mAlphSize, 1, 1, 0, stdLen);

		// Read direction
		dir = RuleBoth;
		if(*bp == '<')
			++bp; // skip <
		else
			dir = RuleToRight;
		++bp; // skip -
		if(*bp == '>')
			++bp; // skip >
		else
			dir = RuleToLeft;

		// Read right itemset
		stdLen = 0.0;
		stdLengths = mDDB->GetRightDatabase()->GetStdLengths();
		for(uint32 i=0; i<setLenR; i++) {
			++bp; // skip space
			setR[i] = atoi(bp);
			stdLen += stdLengths[setR[i]];
			while(*bp != ' ' && *bp != '\n' && *bp != '\0')
				++bp;
		}
		isR = ItemSet::Create(mDDB->GetDataType(), setR, setLenR, mAlphSize, 1, 1, 0, stdLen);

		// Create and add the rule
		mRules->push_back(new TranslationRule(isL, dir, isR));
	}

	// Roundup
	delete[] setL;
	delete[] setR;
	delete[] buffer;
	fclose(file);
}
TranslationTable *TranslationTable::LoadTranslationTable(const string &filename, DoubleDatabase *ddb) {
	printf_s("Loading TT: %s\n", filename.c_str());
	TranslationTable *tt = new TranslationTable(ddb);
	tt->ReadFromDisk(filename);
	tt->InitTranslation();
	return tt;
}
