#ifndef __TRANSLATIONTABLE_H
#define __TRANSLATIONTABLE_H

#define TTFILE_BUFFER_LENGTH 100000

class DoubleDatabase;
class Uint16ItemSet;

#define LEFT 0
#define RIGHT 1

struct TranslationStats {
	double		encLeftSize;
	double		encRightSize;
	double		encTableSize;
	double		encSize;
	uint32		numRules;
	uint32		nbUncoveredL;
	uint32		nbErrorsL;
	uint32		nbUncoveredR;
	uint32		nbErrorsR;

};

#include <Bass.h>
#include "../bass/itemstructs/Uint16ItemSet.h"
#include "TranslationRule.h"

class TranslationTable {
public:
	TranslationTable(DoubleDatabase *ddb);
	TranslationTable(const TranslationTable &tt);
	virtual ~TranslationTable();
	virtual TranslationTable* Clone();

	/* ---- Basic get/set ---- */
	uint32				GetNumRules()				{ return (uint32) mRules->size(); }
	rulelist*				GetRules()				{ return mRules; }

	TranslationStats&	GetCurStats()				{ return mCurStats; }
	double				GetEncodedSize()			{ return mCurStats.encSize; }

	uint32				GetAlphabetSize()			{ return mAlphSize; }
	uint32				GetAlphabetSplit()			{ return mAlphSplit; }

	ItemSet*				GetLeftCovered()			{ return mLeftItems; }
	ItemSet*				GetRightCovered()			{ return mRightItems; }

	ItemSet*				GetLeftLast()			{ return mLeftItemsLast; }
	ItemSet*				GetRightLast()			{ return mRightItemsLast; }
	void                            ClearLast();

	/* ---- Modify the rule set ---- */
	// Permanently adds a rule to the table and applies it to obtain fresh encoded sizes
	void				Add(TranslationRule *rule);
	// Compute the gain that the given rule would give if it were added; does not influence TT state
	double				ComputeGainForAddition(TranslationRule *rule);
	// Compute the gain that the would be achieved by adding a rule base on the given sets; does not influence TT state
	double				ComputeGainForAddition(ItemSet *leftSet, ItemSet *rightSet, double &toRight, double &toLeft);
	double				ComputeGainForAdditionMS(ItemSet *leftSet, ItemSet *rightSet, double &toRight, double &toLeft);
	// Alternative way to compute the gain
	double				ComputeGainForAdditionMod(ItemSet *leftSet, ItemSet *rightSet, ItemSet *leftTidSet, ItemSet *rightTidSet, double &toRight, double &toLeft);
	// For use by RuleTree (note: this one returns `raw' gains, without taking rule sizes into account)
	void				ComputeGainForAdditionFromTree(ItemSet *leftSet, ItemSet *rightSet, Uint16ItemSet *leftTidSet, Uint16ItemSet *rightTidSet, double &toRight, double &toLeft);
	void				ComputeGainForAdditionFromTreeMS(ItemSet *leftSet, ItemSet *rightSet, Uint16ItemSet *leftTidSet, Uint16ItemSet *rightTidSet, double &toRight, double &toLeft);

	/* --- Some useful computations --- */
	float**				ComputeRowGainPotentials();


	/* ---- Perform translation ---- */
	void				InitTranslation();			// build structures required to do translation
	void				Translate()					{ PerformTranslation(true, true); }
	void				TranslateToLeft()			{ PerformTranslation(false, true); }
	void				TranslateToRight()			{ PerformTranslation(true, false); }

	void				ComputeEncodedSize();

	/* ---- Load/save ---- */
	virtual void		WriteToDisk(const string &filename);
	virtual void		ReadFromDisk(const string &filename);
	static TranslationTable* LoadTranslationTable(const string &filename, DoubleDatabase *ddb);

protected:
	/* Helpers */
	void				ResetTranslation();			// resets the correction table to initial position, so that a full new translation can be done
	void				PerformTranslation(const bool ltr, const bool rtl); // does a full new translation, starting from scratch
	void				ApplyRule(TranslationRule *rule, const bool ltr = true, const bool rtl = true);
	void				ApplyRule(ItemSet *antecedent, ItemSet **antRows, ItemSet *consequent, ItemSet **consRows);

	/* Variables */
	DoubleDatabase*		mDDB;
	rulelist*			mRules;

	TranslationStats	mPrevStats;
	TranslationStats	mCurStats;
	//TranslationStats	mPrunedStats;

	/* Translation structures */
	uint32				mNumRows;
	ItemSet				**mUncovered;
	ItemSet				**mErrors;

	ItemSet				*mLeftItems;
	ItemSet				*mRightItems;
	ItemSet				*mLeftItemsLast;
	ItemSet				*mRightItemsLast;
	uint32*				mColUncoveredCounts;
	uint32*				mColErrorCounts;
	uint32*				mGainCounts;
	uint32*				mZeroCounts;

	uint32				mAlphSize, mAlphSplit;
	Database*		mVerticalDb;
	Uint16ItemSet**	mItems;
	std::map<string, Uint16ItemSet*>      mSetsLMap;
	std::map<string, Uint16ItemSet*>      mSetsRMap;

	uint16*				mTmpValues;
	double				*mLeftStdLens;
	double				*mRightStdLens;

	friend class RuleNode;
	friend class RuleNodeCache;
};

#endif // __TRANSLATIONTABLE_H
